@@ -1,6 +1,8 @@
 # almoststatic
+from audioop import reverse
 import datetime
 import os
+import copy
 import glob
 import time
 import json
@@ -22,9 +24,8 @@ class Almoststatic():
     """
 
     # list of names of metadata recognized for each page
-    meta_names = ['filename', 'pagename', 'title', 'description',
-                  'author', 'tags', 'templ_name', 'extends', 'cacheable',
-                  'date']
+    meta_names = ['filename', 'pagename', 'title', 'description', 'author',
+                  'templ_name', 'extends', 'cacheable', 'date', 'blog']
 
     def __init__(self):
 
@@ -79,6 +80,9 @@ class Almoststatic():
         It is mandatory to use it at least once. It can update default
         configuration values and config page globals.
 
+        During development config file can change often, so can be useful to
+        call it each time the page is reloaded.
+
         It also configure the environment for Jinja2 and setup some utilities
         that can be used in templates.
 
@@ -99,9 +103,10 @@ class Almoststatic():
         self.static_url = config['config'].get("static_url", self.static_url)
         self.cache = config['config'].get("cache", self.cache)
 
-        # set the environmet and add globals
+        # set the environment and add globals
         self.env = Environment(
-            loader=FileSystemLoader(self.templates)
+            loader=FileSystemLoader(self.templates),
+            autoescape=False
         )
         # set globals shortcut
         self.vars = self.env.globals
@@ -121,11 +126,12 @@ class Almoststatic():
                 writing static pages this change to site url."""
 
         self.vars['url_suffix'] = ""
-        """str: suffic for url's by default is empty, but during writing static
+        """str: suffix for url's by default is empty, but during writing static
                 pages it will be changed to ".html" """
 
-        # functions addedd to templates
-        self.vars['enum_list'] = enum_list
+        # functions added to templates
+        self.vars['enum_list'] = enumerate
+        self.vars['len_list'] = len
         self.vars['iif'] = iif
         self.vars['if_in'] = if_in
         self.vars['get_markdown'] = get_markdown
@@ -133,6 +139,8 @@ class Almoststatic():
         self.vars['include'] = self._import_file
         self.vars['embed'] = self.render_content
         self.vars['get_url'] = self.get_url
+        self.vars['query_blog'] = self.query_blog
+        self.vars['pages_id'] = self.pages_id
 
     def build_page(self, pagename, **kwargs):
         """This is the main method used to render pages. The pagename is the
@@ -150,13 +158,21 @@ class Almoststatic():
             **kwargs: optional parameters used to pass dynamic contents
 
         Returns:
-            str: the text of rendred page
+            str: the text of rendered page
         """
 
-        # Store metadata of pages into global vars to be accessible during the
-        # page rendering. This can be useful for collecting pages by metadata
+        # Store metadata of pages into global vars and save them in json file
+        # to be accessible during the page rendering and on client side by
+        # scripts. This can be useful for collecting pages by metadata
         if not self.vars.get('meta_pages', None):
             self.vars['meta_pages'] = self.meta_pages()
+            filename = os.path.join(
+                self.content, self.media, 'site_meta.json')
+            jmeta = json.dumps(self.vars['meta_pages'],
+                               indent=4, default=_json_date_encoder)
+            with open(filename, "w") as f:
+                f.write(jmeta)
+                f.close()
 
         filename, page = self.load_yaml_page(pagename)
         if not page:
@@ -177,7 +193,8 @@ class Almoststatic():
             self.extra_args = kwargs
             self.page = page
             templ_name = page.get('templ_name', None)
-            content = self.render_content(page['content'])
+            content = self.render_content(
+                page['content'], page.get('widgets_envelope'))
             content = self.render_template(
                 templ_name=templ_name, content=content)
             # template commands can be embedded within html and md and this is
@@ -189,7 +206,7 @@ class Almoststatic():
         return content
 
     def write_static(self, pages=None, destination='.', media_prefix=None,
-                     static_url=None):
+                     static_url=None, out_pages=[]):
         """Write html pages on disk! If pages is None, all pages declared in
         content/pages folder are rendered as html and written on disk.
 
@@ -216,6 +233,7 @@ class Almoststatic():
                 prefix for local links, so if the destination is
                 https://www.example.com/mysite with prefix /mysite/ all links
                 works fine
+            out_pages (list): list of pages excluded from writing
         """
         self.vars['is_static'] = True
         self.vars['url_suffix'] = ".html"
@@ -232,6 +250,8 @@ class Almoststatic():
             pages = [pages]
 
         for page in pages:
+            if page in out_pages:
+                continue
             s = self.build_page(page)
             filename = os.path.join(
                 destination, page+self.env.globals['url_suffix'])
@@ -249,6 +269,11 @@ class Almoststatic():
         """Write metadata of pages as json file to disk. So if needed
         javascript can read them to build client side logics.
 
+        Note:
+            This method usually is not needed, because each time Almoststatic
+            starts and the first page is rendered, metadata are written on:
+            ./content/media/site_meta.json so metadata can be found there.
+
         Args:
             filename (list): Name of json file to write
             destination (str): destination path folder for json file
@@ -271,24 +296,24 @@ class Almoststatic():
         Start point of media files usually is: ./content/media
 
         Args:
-            media_obj (str): media obget to retrive with the local path from
+            media_obj (str): media object to retrive with the local path from
                              media starting point
 
         Returns:
-            str: the pathname of media pbject
+            str: the pathname of media object
         """
         if self.media_prefix is None:
-            media_prefix = os.path.join(
-                self._relative_url(), self.content, self.media)
+            media_prefix = '/'.join([self._relative_url(),
+                                     self.content, self.media])
         else:
-            media_prefix = self._relative_url() + self.media_prefix
-        filename = os.path.join(media_prefix, media_obj)
+            media_prefix = self._relative_url() + self.media
+        filename = '/'.join([media_prefix, media_obj])
         return filename
 
     def get_url(self, page):
         """Build the url for local links adding prefix and suffix to the page,
-        used inside templates write local url's consistents in dynamic and
-        static sites
+        used inside templates write local url's consistent in dynamic and
+        static sites. Recognize the anchor separator to add suffix before it.
 
         Args:
             page (str): local url
@@ -296,7 +321,14 @@ class Almoststatic():
         Returns:
             str: the url for internal link
         """
-        return f"{self._relative_url()}{page}{self.vars['url_suffix']}"
+        try:
+            i = page.index('#')
+            anchor = page[i:]
+            page = page[:i]
+        except ValueError:
+            anchor = ''
+        suffix = self.vars['url_suffix']
+        return f"{self._relative_url()}{page}{suffix}{anchor}"
 
     def _relative_url(self):
         """"""
@@ -307,7 +339,7 @@ class Almoststatic():
     def meta_pages(self, subfolder=''):
         """Read metadata of all pages declared in ./content/pages folder. The
         folder parameter is used to read only a sub folder, this can be
-        useful for collecting metadada of only one folder i.e. "blog"
+        useful for collecting metadata of only one folder i.e. "blog"
 
         Args:
             subfolder (str): if present limit metadata reading to the folder
@@ -316,15 +348,109 @@ class Almoststatic():
         """
         files, pages = self.get_site_pages(subfolder)
         meta = {}
-        for pagename in pages:
+        for i, pagename in enumerate(pages):
             filename, page = self.load_yaml_page(pagename)
             m = {}
             for k, v in page.items():
                 if k in self.meta_names:
                     m[k] = v
+            m['id'] = i
             meta[m['pagename']] = m
         return meta
 
+    def query_blog(self, categories=[], tags=[], empty_ok=False):
+        """Query blog pages against categories and tags. Each page can have
+        many tags and belong to many categories. This method query pages by
+        list of categories and list of tags and return the list of pages
+        which matches all criteria and pages for categories and tags. Usually
+        pages with no tags or no categories are not accepted, but if parameter
+        empty_ok is true are accepted at least in list of all pages.
+
+        Args:
+            categories (list): list of categories to search, if empty accept
+                all
+            tags (list): list of tags to search, if empty accept all
+            empty_ok (bool): if true accept pages without categories or tags
+
+        Returns:
+            tuple: list of all pages, list of pages matching categories, list
+            of pages matching tags
+        """
+        pages = []
+        cat_dict = {}
+        tag_dict = {}
+        for k, v in self.vars['meta_pages'].items():
+            if 'blog' not in v:
+                continue
+
+            def _check(lquery, ldata):
+                """check if page belongs to categories or tags in query. return
+                a list with found keys
+                """
+                dest = []
+                ll = copy.copy(lquery) if lquery else ldata
+                for i in ll:
+                    try:
+                        ldata.index(i)
+                        dest.append(i)
+                    except ValueError:
+                        pass
+                return dest
+
+            def _add_page(keys, dest):
+                """Add al keys in det dict and append the page to all"""
+                for i in keys:
+                    if i not in dest.keys():
+                        dest[i] = []
+                    dest[i].append(v)
+
+            cat = _check(categories, v['blog'].get('categories', []))
+            tag = _check(tags, v['blog'].get('tags', []))
+            # accept pages with empty categories or tags, or not
+            if categories or empty_ok:
+                _add_page(cat, cat_dict)
+            if tags or empty_ok:
+                _add_page(tag, tag_dict)
+            if categories or tags or empty_ok:
+                pages.append(v)
+
+        # order categories and tag first by priority and then by date
+        for k in cat_dict.keys():
+            def _sort(page):
+                if page['blog'].get('categ_order'):
+                    t = page['blog']['categ_order'].get(k, 0)
+                else:
+                    t = 0
+                x = f"{t:0>10} {page['date']}"
+                return x
+            cat_dict[k].sort(key=_sort, reverse=True)
+        for k in tag_dict.keys():
+            def _sort(page):
+                if page['blog'].get('tag_order'):
+                    t = page['blog']['tag_order'].get(k, 0)
+                else:
+                    t = 0
+                x = f"{t:0>10} {page['date']}"
+                return x
+            tag_dict[k].sort(key=_sort, reverse=True)
+
+        return (pages, cat_dict, tag_dict)
+
+    def pages_id(self, pages):
+        """Return a list of the id's of pages passed as argument. This is
+        useful to give list of pages to scripts at runtime that can be searched
+        in file './content/media/site_meta.json'
+
+        Args:
+            pages (list): list pages metadata
+        Returns:
+            list: list of id's of passed pages
+        """
+        dest = []
+        for page in pages:
+            dest.append(page['id'])
+        return dest
+
     def get_site_pages(self, subfolder=''):
         """Read the pages folder and return the list of yaml files and the list
         of pages url's
@@ -332,7 +458,7 @@ class Almoststatic():
         Args:
             subfolder (str): used to restrict reading to only a sub folder.
         Returns:
-            tupla: list of files and list of page url's
+            tuple: list of files and list of page url's
         """
         base_path = os.path.join(self.content, self.pages, '')
         path = os.path.join(base_path, subfolder, '')
@@ -354,16 +480,16 @@ class Almoststatic():
         title = title of page
         description = description of page
         author = author of page
-        tags = list of arbitrary tags for page
         templ_name = template name used to render the page
         extends = base page to extend
         cacheable = if page is cacheable it is rendered only once
         date = date of page if not present takes the modification time of file
+        blog = blog data of page if present
 
         Args:
             pagename (str): name of page to read
         Returns:
-            tupla: the filename of page and the dict with page data and
+            tuple: the filename of page and the dict with page data and
                 metadata
         """
         filename = self.get_filename(pagename)
@@ -382,7 +508,6 @@ class Almoststatic():
         default_meta('title', '')
         default_meta('description', '')
         default_meta('author', '')
-        default_meta('tags', [])
 
         page['templ_name'] = page.get('templ_name', self.templ_name)
         page['extends'] = page.get('extends', None)
@@ -395,9 +520,10 @@ class Almoststatic():
         """return the yaml file from pagename"""
         return os.path.join(self.content, self.pages, pagename+'.yaml')
 
-    def render_content(self, content):
+    def render_content(self, content, widgets_envelope=None):
         """Render the content of page, the content is a a key of the page and
-        can be a single widget or a list of widgets
+        can be a single widget or a list of widgets, widgets_envelope is the
+        default evelope for each widgets which can be passed by page
         """
         if not isinstance(content, list):
             content = [content]
@@ -409,10 +535,15 @@ class Almoststatic():
             if widget['type'] == 'include':
                 result += self._import_file(widget['file'])
             else:
-                widget = self.eval_widget(widget)
-                result += self.render_template(
-                    widget['type']+'.html', widget=widget)
-
+                disabled = widget.get('disabled', False)
+                if not disabled:
+                    widget = self.eval_widget(widget)
+                    text = self.render_template(
+                        widget['type']+'.html', widget=widget)
+                    envelope = widget.get('envelope') or widgets_envelope
+                    if envelope:
+                        text = "%s\n%s\n%s" % (envelope[0], text, envelope[1])
+                    result += text
         return result
 
     def eval_widget(self, value, key=''):
@@ -442,15 +573,17 @@ class Almoststatic():
     def _import_file(self, filename):
         """Allow the inclusion of files containg widgets or plain html or
         markdown files. Nested inclusions are allowed and the limit is only
-        the python call stack, so you have only to avoid circular inculsions.
-        the recognized extensions are: '.html' for plain html text, '.md' for
-        markdown text and '.yaml' which can contain widgets under 'content' key
+        the python call stack, so you have only to avoid circular inclusions.
+        the recognized extensions are: '.html' or '.htm' for plain html text,
+        '.md' for markdown text and '.yaml' which can contain widgets under
+        'content' key
         """
         fname = os.path.join(self.content, filename)
         if not os.path.isfile(fname):
-            return 'The file: %s does not exists.' % filename
+            return f'The file: {filename} does not exists.'
         with open(fname) as f:
-            if fname.endswith('.html'):
+            s = f"File: {filename} not recognized type."
+            if fname.endswith('.html') or fname.endswith('.htm'):
                 s = f.read()
             elif fname.endswith('.md'):
                 s = f.read()
@@ -460,7 +593,6 @@ class Almoststatic():
                 s = self.render_content(content)
             f.close()
             return s
-        return "File: %s not recognized type." % filename
 
     def render_template(self, templ_name=None, **kwargs):
         """Render the template and return the text rendered"""
@@ -478,7 +610,6 @@ class FlaskAlmoststatic(Almoststatic):
 
     Args:
         app (flask.app): the flask app where Almoststtic is attached.
-
     """
 
     def __init__(self, app=None):
@@ -499,6 +630,8 @@ class FlaskAlmoststatic(Almoststatic):
         rule = f'/{self.content}/<path:filename>'
         app.add_url_rule(rule, 'content', self.content_static)
         app.template_context_processors[None].append(self.content_processor)
+        app.jinja_env.autoescape = False
+        app.jinja_loader = FileSystemLoader(self.templates)
 
     def content_processor(self):
         return self.vars
@@ -518,17 +651,6 @@ class FlaskAlmoststatic(Almoststatic):
         return s
 
 
-def enum_list(list_to_enum):
-    """add enumerate function to jinja2
-
-    Args:
-        list_to_enum (list): the list to enumerate
-    Returns:
-        tuple: the enumerated list
-    """
-    return enumerate(list_to_enum)
-
-
 def get_markdown(text):
     """convert markdown to html and render tables
 
@@ -541,7 +663,7 @@ def get_markdown(text):
 
 
 def iif(condition, if_true, if_false=""):
-    """Inline if shortcode
+    """Inline if short code
     try: iif(1==1, "Ok", "No")
 
     Args:
@@ -558,7 +680,7 @@ def iif(condition, if_true, if_false=""):
 
 def if_in(value, iterable, if_not=""):
     """Inline check for inclusion.
-    try: ifin('a', ['a','b','c'], 'd')
+    try: if_in('a', ['a','b','c'], 'd')
 
     Args:
         value (any): value to check
@@ -577,5 +699,5 @@ def if_in(value, iterable, if_not=""):
 
 def _json_date_encoder(value):
     """encoder for json dumps of dates"""
-    if isinstance(value, datetime.datetime):
+    if isinstance(value, datetime.date):
         return value.__str__()
